from .run import run_pauli_expectation
from .resolution import build_quasar_circuit
from .circuit import Circuit
import numpy as np
import itertools

class Tomography(object):

    def __init__(self):
        raise NotImplementedError

    @property
    def nparam(self):
        raise NotImplementedError
        

    def compute_observable_expectation_value(
        self,
        params, 
        ):
        raise NotImplementedError

    def compute_observable_expectation_value_gradient(
        self,
        params,
        ):

        raise NotImplementedError

    def compute_observable_expectation_value_hessian(
        self,
        params,
        ):

        raise NotImplementedError

class RotationTomography(Tomography):

    def __init__(
        self,
        coefs,
        ):

        self.coefs = coefs

    @property
    def nparam(self):
        return self.coefs.ndim
        
    def compute_observable_expectation_value(
        self,
        params,
        ):
    
        if params.ndim < 2: raise RuntimeError('params.ndim < 2')
        if params.shape[0] != self.nparam: raise RuntimeError('params.shape[0] != self.nparam')
    
        # NOTE: Special case (bare coefficent)
        if self.coefs.ndim == 0: return self.coefs
    
        bs = []
        for theta in params:
            bs.append([
                np.ones_like(theta),
                np.cos(2.0 * theta),
                np.sin(2.0 * theta),  
                ])
    
        O = np.zeros_like(params[0])
        for Js in itertools.product(range(3), repeat=self.nparam):
            R = np.ones_like(params[0])
            for J2, J in enumerate(Js):
                R *= bs[J2][J]
            O += self.coefs[Js] * R
    
        return O
    
    def compute_observable_expectation_value_gradient(
        self,
        params,
        ):
    
        if params.ndim < 2: raise RuntimeError('params.ndim < 2')
        if params.shape[0] != self.nparam: raise RuntimeError('params.shape[0] != self.nparam')
    
        G = np.zeros((self.nparam,) + params[0].shape)
        for k in range(self.nparam):
            notk = [_ for _ in range(self.nparam) if _ != k] 
            paramsk = params[k]
            params2 = params[notk]
            coefs_b = np.take(self.coefs, 1, k)
            coefs_c = np.take(self.coefs, 2, k)
            tomography_b = RotationTomography(coefs_b)
            tomography_c = RotationTomography(coefs_c)
            Ob = tomography_b.compute_observable_expectation_value(params2)
            Oc = tomography_c.compute_observable_expectation_value(params2)
            G[k] = -2.0 * np.sin(2.0 * paramsk) * Ob + 2.0 * np.cos(2.0 * paramsk) * Oc
        return G
    
    def compute_observable_expectation_value_hessian(
        self,
        params,
        ):
    
        if params.ndim < 2: raise RuntimeError('params.ndim < 2')
        if params.shape[0] != self.nparam: raise RuntimeError('params.shape[0] != self.nparam')
    
        H = np.zeros((self.nparam,)*2 + params[0].shape)
        for k in range(self.nparam):
            notk = [_ for _ in range(self.nparam) if _ != k] 
            paramsk = params[k]
            params2 = params[notk]
            coefs_b = np.take(self.coefs, 1, k)
            coefs_c = np.take(self.coefs, 2, k)
            tomography_b = RotationTomography(coefs_b)
            tomography_c = RotationTomography(coefs_c)
            Ob = tomography_b.compute_observable_expectation_value(params2)
            Oc = tomography_c.compute_observable_expectation_value(params2)
            if len(notk):
                Gb = tomography_b.compute_observable_expectation_value_gradient(params2)
                Gc = tomography_c.compute_observable_expectation_value_gradient(params2)
                H[k,notk] = -2.0 * np.sin(2.0 * paramsk) * Gb + 2.0 * np.cos(2.0 * paramsk) * Gc
            H[k, k] = - 4.0 * np.cos(2.0 * paramsk) * Ob - 4.0 * np.sin(2.0 * paramsk) * Oc
        return H 

    # > Tomography quadrature utility < #

    @staticmethod
    def quad_x(D=1):
    
        return np.array(np.meshgrid(
            *[[-np.pi / 3.0, 0.0, +np.pi / 3.0]]*D,
            indexing='ij',
            ))
    
    @staticmethod
    def quad_transfer(D=1):
    
        T1 = np.array([
            [1.0, -0.5, -np.sqrt(3.0)/2.0,],
            [1.0,  1.0,               0.0,],
            [1.0, -0.5, +np.sqrt(3.0)/2.0,],
            ], dtype=np.float)
    
        T = np.copy(T1)
        for D2 in range(1,D):
            T = np.kron(T, T1)
        T *= 0.5**D
    
        return T
    
    @staticmethod
    def quad_transfer_inv(D=1):
    
        T1inv = np.array([
            [         1.0,  1.0,           1.0,],
            [        -1.0,  2.0,          -1.0,],
            [-np.sqrt(3.0), 0.0, +np.sqrt(3.0),],
            ], dtype=np.float) / 3.0
    
        Tinv = np.copy(T1inv)
        for D2 in range(1,D):
            Tinv = np.kron(Tinv, T1inv)
    
        return Tinv
    
    @staticmethod
    def quad_coefs(O):
    
        return np.reshape(np.dot(RotationTomography.quad_transfer_inv(D=O.ndim), O.ravel()), O.shape)

    # => Optimization <= #
    
    def optimize_jacobi_1(
        self,
        theta0=None,
        n=100,
        d=0,
        ):
    
        if theta0 is None:
            theta0 = np.zeros((self.nparam,))
    
        theta = np.copy(theta0)
        thetas = [theta0]
        for iteration in range(n):
            k = (iteration + d) % self.nparam
            theta_2 = np.array([theta for k2, theta in enumerate(theta) if k2 != k])
            theta_2 = np.reshape(theta_2, theta_2.shape + (1,))
            coefs_b = np.take(self.coefs, 1, k)
            coefs_c = np.take(self.coefs, 2, k)
            tomography_b = RotationTomography(coefs_b)
            tomography_c = RotationTomography(coefs_c)
            Ob = tomography_b.compute_observable_expectation_value(theta_2)
            Oc = tomography_c.compute_observable_expectation_value(theta_2)
            theta[k] = 0.5 * np.arctan2(-Oc, -Ob)
            thetas.append(np.copy(theta))
        thetas = np.array(thetas)
        return thetas.T
    
    def optimize_jacobi_1_best(
        self,
        theta0=None,
        n=100,
        ):
    
        thetas = []
        for d in range(self.nparam):
            thetas.append(self.optimize_jacobi_1(
                theta0=theta0,  
                n=n,    
                d=d,
                ))
        Os = np.array([self.compute_observable_expectation_value(theta2[:,-2:-1]) for theta2 in thetas])
        return thetas[np.argmin(Os)] 

    def optimize(self):
        return self.optimize_jacobi_1_best()[:,-1]

def run_observable_expectation_value_tomography(
    backend,
    circuit,
    pauli,
    nmeasurement=None,
    param_indices=None,
    **kwargs):
    
    # No dropthrough - always need quasar.Circuit to manipulate
    circuit = build_quasar_circuit(circuit).copy()
    param_values = circuit.param_values

    # Default to doing tomography over all parameters (NOTE: This costs 3**nparam pauli expectation values)
    if param_indices is None:
        param_indices = tuple(range(circuit.nparam))

    # Check that the tomography formula is known for these parameters (i.e., Rx, Ry, Rz gates)
    param_keys = circuit.param_keys
    for param_index in param_indices:
        key = param_keys[param_index]
        time, qubits, name = key
        gate = circuit.gates[(time, qubits)]
        if not gate.name in ('Rx', 'Ry', 'Rz'): 
            raise RuntimeError('Unknown tomography rule: presently can only tomography Rx, Ry, Rz gates: %s' % gate)

    # The tomography quadrature grid
    T = RotationTomography.quad_x(len(param_indices))
    O = np.zeros_like(T[0])
    npoint = T.size // T.shape[0]
    for I in range(npoint):
        param_values2 = param_values.copy()
        for param_index, T2 in zip(param_indices, T):
            param_values2[param_index] = T2.ravel()[I]
        circuit.set_param_values(param_values2)
        O.ravel()[I] = run_pauli_expectation(backend, circuit, pauli, nmeasurement, **kwargs).dot(pauli).real # TODO: do we need this to be real
    
    # Tomography fitting
    coefs = RotationTomography.quad_coefs(O) 

    # Finished RotationTomography object
    return RotationTomography(coefs=coefs) 

def run_ensemble_observable_expectation_value_tomography(
    backend,
    reference_circuits,
    reference_weights,
    circuit,
    pauli,
    nmeasurement=None,
    param_indices=None,
    **kwargs):
    
    # No dropthrough - always need quasar.Circuit to manipulate
    reference_circuits = [build_quasar_circuit(_) for _ in reference_circuits]
    circuit = build_quasar_circuit(circuit).copy()
    param_values = circuit.param_values

    # Default to doing tomography over all parameters (NOTE: This costs 3**nparam pauli expectation values)
    if param_indices is None:
        param_indices = tuple(range(circuit.nparam))

    # Check that the tomography formula is known for these parameters (i.e., Rx, Ry, Rz gates)
    param_keys = circuit.param_keys
    for param_index in param_indices:
        key = param_keys[param_index]
        time, qubits, name = key
        gate = circuit.gates[(time, qubits)]
        if not gate.name in ('Rx', 'Ry', 'Rz'): 
            raise RuntimeError('Unknown tomography rule: presently can only tomography Rx, Ry, Rz gates: %s' % gate)

    # The tomography quadrature grid
    T = RotationTomography.quad_x(len(param_indices))
    O = np.zeros_like(T[0])
    npoint = T.size // T.shape[0]
    for I in range(npoint):
        param_values2 = param_values.copy()
        for param_index, T2 in zip(param_indices, T):
            param_values2[param_index] = T2.ravel()[I]
        circuit.set_param_values(param_values2)
        Oval = 0.0
        for ref, w in zip(reference_circuits, reference_weights):
            circuit2 = Circuit.concatenate([ref, circuit])
            Oval += w * run_pauli_expectation(backend, circuit2, pauli, nmeasurement, **kwargs).dot(pauli).real # TODO: do we need this to be real
        O.ravel()[I] = Oval
    
    # Tomography fitting
    coefs = RotationTomography.quad_coefs(O) 

    # Finished RotationTomography object
    return RotationTomography(coefs=coefs) 






import numpy as np
from .backend import Backend
from .circuit import Gate, Circuit

class QuasarSimulatorBackend(Backend):

    def __init__(
        self,
        ):
        pass

    def __str__(self):
        return 'Quasar Simulator Backend (Statevector)'

    @property
    def summary_str(self):
        s = ''
        s += 'Quasar: An Ultralite Quantum Circuit Simulator\n'
        s += '   By Rob Parrish (rob.parrish@qcware.com)    '
        return s

    @property
    def has_run_statevector(self):
        return True

    @property
    def has_statevector_input(self):
        return True

    def run_statevector(
        self,
        circuit,
        statevector=None,   
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        ):

        min_qubit = circuit.min_qubit if min_qubit is None else min_qubit
        nqubit = circuit.nqubit if nqubit is None else nqubit

        if statevector is None:
            statevector1 = np.zeros((2**nqubit,), dtype=dtype)
            statevector1[0] = 1.0
        else:
            statevector1 = statevector.copy()
        statevector2 = np.zeros_like(statevector1)

        qubits = [_ - min_qubit for _ in range(circuit.min_qubit, circuit.min_qubit + circuit.nqubit)]

        return circuit.apply_to_statevector(
            statevector1=statevector1,
            statevector2=statevector2,
            qubits=qubits,
            dtype=dtype,
            )[0]

    def run_pauli_sigma(
        self,
        pauli,
        statevector,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        min_qubit = pauli.min_qubit if min_qubit is None else min_qubit
        nqubit = pauli.nqubit if nqubit is None else nqubit

        pauli_gates_mody = {
            'X' : Gate.X,
            'Y' : Gate.U1(np.array([[0.0, 1.0], [-1.0, 0.0]], dtype=np.complex128)),
            'Z' : Gate.Z,
        }

        statevector2 = np.zeros_like(statevector)
        for string, value in pauli.items():
            circuit2 = Circuit()
            ny = 0
            for qubit, char in string:
                circuit2.add_gate(pauli_gates_mody[char], qubit)
                if char == 'Y':
                    ny += 1
            statevector3 = self.run_statevector(
                circuit=circuit2,
                statevector=statevector,
                dtype=dtype,
                min_qubit=min_qubit,
                nqubit=nqubit,
                )
            scal = (-1.j)**ny * value
            if dtype in (np.float32, np.float64):
                scal = dtype(scal.real)
            else:
                scal = dtype(scal)
            statevector2 += scal * statevector3

        return statevector2




import numpy as np
import collections
import sortedcontainers
from .pauli import Pauli, PauliExpectation, PauliString
from .algebra import Algebra
from .circuit import Gate, Circuit

class Backend(object):

    """ Class Backend represents a physical or simulated quantum circuit
        resource. Backends must implement `run_measurement`, from which many
        higher-order quantities may be computed, e.g., `run_pauli_expectation`,
        `run_pauli_expectation_value`, `run_pauli_expectation_value_gradient`,
        etc. Backends supporting classical statevector-based simulation may
        also optionally implement `run_statevector,` from which many additional
        higher-order quantities may be computed, e.g., `run_unitary`,
        `run_density_matrix`, and ideal-infinite-sampled versions of the
        previously-discussed higher-order quantities. Backends may additionally
        overload any of the stock higher-order methods declared here,
        potentially providing increased performance or additional methodology.
    """ 

    def __init__(
        self,
        ):

        """ Constructor, initializes and holds quantum resource pointers such
            as API keys.

        Backend subclasses should OVERLOAD this method.
        """
        pass

    def __str__(self):
        """ A 1-line string representation of this Backend

        Returns:
            (str) - 1-line string representation of this Backend
        
        Backend subclasses should OVERLOAD this method.
        """ 
        raise NotImplementedError

    @property
    def summary_str(self):
        """ A more-extensive string representation of this Backend, optionally
            including current hardware state.

        Returns:
            (str) - multiline string representation of this Backend
        
        Backend subclasses should OVERLOAD this method.
        """ 
        raise NotImplementedError

    @property
    def has_run_statevector(
        self,
        ):

        """ Does this Backend support run_statevector? 

        Returns:
            (bool) - True if run_statevector is supported else False.

        Backend subclasses should OVERLOAD this method.
        """ 
        raise NotImplementedError

    @property
    def has_statevector_input(
        self,
        ):

        """ Does this Backend allow statevector to be passed as input argument
            to various run methods?

        Returns:
            (bool) - True if statevector input arguments can be supplied else
                False.

        Backend subclasses should OVERLOAD this method.
        """

        raise NotImplementedError

    def run_statevector(
        self,
        circuit,
        statevector=None,   
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        raise NotImplementedError

    def run_pauli_sigma(
        self,
        pauli,
        statevector,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        min_qubit = pauli.min_qubit if min_qubit is None else min_qubit
        nqubit = pauli.nqubit if nqubit is None else nqubit

        pauli_gates = {
            'X' : Gate.X,
            'Y' : Gate.Y,
            'Z' : Gate.Z,
        }

        statevector2 = np.zeros_like(statevector)
        for string, value in pauli.items():
            circuit2 = Circuit()
            for qubit, char in string:
                circuit2.add_gate(pauli_gates[char], qubit)
            statevector3 = self.run_statevector(
                circuit=circuit2,
                statevector=statevector,
                dtype=dtype,
                min_qubit=min_qubit,
                nqubit=nqubit,
                )
            statevector2 += value * statevector3

        return statevector2

    def run_pauli_diagonal(
        self,
        pauli,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        min_qubit = pauli.min_qubit if min_qubit is None else min_qubit
        nqubit = pauli.nqubit if nqubit is None else nqubit

        # All I or Z strings
        pauli2 = Pauli.zero()
        for string, value in pauli.items():
            if len(string) == 0 or all(_ == 'Z' for _ in string.chars):
                pauli2[string] = value

        statevector = np.ones((2**nqubit,), dtype=dtype)
        return self.run_pauli_sigma(
            pauli=pauli2,
            statevector=statevector,
            min_qubit=min_qubit,
            nqubit=nqubit,
            dtype=dtype,
            **kwargs)

    def run_unitary(
        self,
        circuit,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        nqubit = circuit.nqubit if nqubit is None else nqubit

        U = np.zeros((2**nqubit,)*2, dtype=dtype)
        for i in range(2**nqubit):
            statevector = np.zeros((2**nqubit,), dtype=dtype)
            statevector[i] = 1.0
            U[:, i] = self.run_statevector(circuit, statevector=statevector, dtype=dtype, **kwargs)

        return U

    def run_density_matrix(
        self,
        circuit,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        statevector = self.run_statevector(
            circuit=circuit,
            statevector=statevector,
            min_qubit=min_qubit,
            nqubit=nqubit,
            dtype=dtype,
            **kwargs)

        return np.outer(statevector, statevector.conj())

    def run_measurement(
        self,
        circuit,
        nmeasurement=None,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        statevector = self.run_statevector(
            circuit=circuit,
            statevector=statevector,
            min_qubit=min_qubit,
            nqubit=nqubit,
            dtype=dtype,
            **kwargs)

        return Algebra.sample_histogram_from_probabilities(
            probabilities=(np.conj(statevector) * statevector).real,
            nmeasurement=nmeasurement,
            )

    def run_pauli_expectation(
        self,
        circuit,
        pauli,
        nmeasurement=None,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):
        
        if nmeasurement is None:
            return self.run_pauli_expectation_ideal(
                circuit=circuit,
                pauli=pauli,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
        else:
            return self.run_pauli_expectation_measurement(
                circuit=circuit,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)

    def run_pauli_expectation_value(
        self,
        circuit,
        pauli,
        nmeasurement=None,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        if nmeasurement is None:
            return self.run_pauli_expectation_value_ideal(
                circuit=circuit,
                pauli=pauli,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
        else:
            pauli_expectation = self.run_pauli_expectation_measurement(
                circuit=circuit,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
            return pauli_expectation.dot(pauli)

    def run_pauli_expectation_value_gradient(
        self,
        circuit,
        pauli,
        nmeasurement=None,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        parameter_indices=None,
        **kwargs):
        
        # Current circuit parameter values
        parameter_values = circuit.parameter_values

        # Default to taking the gradient with respect to all parameters
        if parameter_indices is None:
            parameter_indices = list(range(len(parameter_values)))

        # Check that the gradient formula is known for these parameters (i.e., Rx, Ry, Rz gates)
        parameter_keys = circuit.parameter_keys
        for parameter_index in parameter_indices:
            key = parameter_keys[parameter_index]
            times, qubits, key2 = key
            gate = circuit.gates[(times, qubits)]
            if not gate.name in ('Rx', 'Ry', 'Rz'): 
                raise RuntimeError('Unknown gradient rule: presently can only differentiate Rx, Ry, Rz gates: %s' % gate)

        # Evaluate the gradient by the parameter shift rule
        gradient = np.zeros((len(parameter_indices),), dtype=dtype)
        circuit2 = circuit.copy()
        for index, parameter_index in enumerate(parameter_indices):
            # +
            parameter_values2 = parameter_values.copy()
            parameter_values2[parameter_index] += np.pi / 4.0
            circuit2.set_parameter_values(parameter_values2)
            Ep = self.run_pauli_expectation_value(
                circuit=circuit2,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
            # -
            parameter_values2 = parameter_values.copy()
            parameter_values2[parameter_index] -= np.pi / 4.0
            circuit2.set_parameter_values(parameter_values2)
            Em = self.run_pauli_expectation_value(
                circuit=circuit2,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
            # Assembly
            gradient[index] = Ep - Em

        return gradient

    def run_pauli_expectation_value_hessian(
        self,
        circuit,
        pauli,
        nmeasurement=None,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        parameter_pair_indices=None,
        **kwargs):

        # Current circuit parameter values
        parameter_values = circuit.parameter_values

        # Default to taking the gradient with respect to all parameters
        if parameter_pair_indices is None:
            parameter_pair_indices = []
            for i in range(len(parameter_values)):
                for j in range(len(parameter_values)):
                    parameter_pair_indices.append((i,j))

        # Check that the Hessian formula is known for these parameters (i.e., Rx, Ry, Rz gates)
        parameter_keys = circuit.parameter_keys
        for parameter_index1, parameter_index2 in parameter_pair_indices:
            key = parameter_keys[parameter_index1]
            times, qubits, key2 = key
            gate = circuit.gates[(times, qubits)]
            if not gate.name in ('Rx', 'Ry', 'Rz'): 
                raise RuntimeError('Unknown Hessian rule: presently can only differentiate Rx, Ry, Rz gates: %s' % gate)
            key = parameter_keys[parameter_index2]
            times, qubits, key2 = key
            gate = circuit.gates[(times, qubits)]
            if not gate.name in ('Rx', 'Ry', 'Rz'): 
                raise RuntimeError('Unknown Hessian rule: presently can only differentiate Rx, Ry, Rz gates: %s' % gate)

        # Evaluate the gradient by the parameter shift rule
        hessian = np.zeros((len(parameter_pair_indices),), dtype=dtype)
        circuit2 = circuit.copy()
        for index, parameter_pair_index in enumerate(parameter_pair_indices):
            parameter_index1, parameter_index2 = parameter_pair_index
            symmetric = (parameter_index2, parameter_index1) in parameter_pair_indices
            if symmetric and parameter_index1 > parameter_index2: continue
            # ++
            parameter_values2 = parameter_values.copy()
            parameter_values2[parameter_index1] += np.pi / 4.0
            parameter_values2[parameter_index2] += np.pi / 4.0
            circuit2.set_parameter_values(parameter_values2)
            Epp = self.run_pauli_expectation_value(
                circuit=circuit2,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
            # +-
            parameter_values2 = parameter_values.copy()
            parameter_values2[parameter_index1] += np.pi / 4.0
            parameter_values2[parameter_index2] -= np.pi / 4.0
            circuit2.set_parameter_values(parameter_values2)
            Epm = self.run_pauli_expectation_value(
                circuit=circuit2,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
            # -+
            parameter_values2 = parameter_values.copy()
            parameter_values2[parameter_index1] -= np.pi / 4.0
            parameter_values2[parameter_index2] += np.pi / 4.0
            circuit2.set_parameter_values(parameter_values2)
            Emp = self.run_pauli_expectation_value(
                circuit=circuit2,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
            # --
            parameter_values2 = parameter_values.copy()
            parameter_values2[parameter_index1] -= np.pi / 4.0
            parameter_values2[parameter_index2] -= np.pi / 4.0
            circuit2.set_parameter_values(parameter_values2)
            Emm = self.run_pauli_expectation_value(
                circuit=circuit2,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
            # Assembly
            hessian[index] = Epp - Epm - Emp + Emm
            if symmetric:
                hessian[parameter_pair_indices.index((parameter_index2, parameter_index1))] = hessian[index]

        return hessian

    def run_pauli_expectation_value_gradient_pauli_contraction(
        self,
        circuit,
        pauli,
        parameter_coefficients,
        nmeasurement=None,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        parameter_indices=None,
        **kwargs):
        
        # Current circuit parameter values
        parameter_values = circuit.parameter_values

        # Default to taking the gradient with respect to all parameters
        if parameter_indices is None:
            parameter_indices = list(range(len(parameter_values)))

        # Check that parameter coefficients make sense
        if len(parameter_coefficients) != len(parameter_indices):
           raise RuntimeError('len(parameter_coefficients) != len(parameter_indices)')

        # Check that the gradient formula is known for these parameters (i.e., Rx, Ry, Rz gates)
        parameter_keys = circuit.parameter_keys
        for parameter_index in parameter_indices:
            key = parameter_keys[parameter_index]
            times, qubits, key2 = key
            gate = circuit.gates[(times, qubits)]
            if not gate.name in ('Rx', 'Ry', 'Rz'): 
                raise RuntimeError('Unknown gradient rule: presently can only differentiate Rx, Ry, Rz gates: %s' % gate)

        # Evaluate the gradient by the parameter shift rule
        pauli_gradient = PauliExpectation.zero()
        circuit2 = circuit.copy()
        for index, parameter_index in enumerate(parameter_indices):
            # +
            parameter_values2 = parameter_values.copy()
            parameter_values2[parameter_index] += np.pi / 4.0
            circuit2.set_parameter_values(parameter_values2)
            Gp = self.run_pauli_expectation(
                circuit=circuit2,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
            # -
            parameter_values2 = parameter_values.copy()
            parameter_values2[parameter_index] -= np.pi / 4.0
            circuit2.set_parameter_values(parameter_values2)
            Gm = self.run_pauli_expectation(
                circuit=circuit2,
                pauli=pauli,
                nmeasurement=nmeasurement,
                statevector=statevector,
                min_qubit=min_qubit,
                nqubit=nqubit,
                dtype=dtype,
                **kwargs)
            # Assembly
            pauli_gradient += parameter_coefficients[index] * (Gp - Gm)

        return pauli_gradient

    # => Utility Methods <= #

    def run_pauli_expectation_ideal(
        self,
        circuit,
        pauli,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        min_qubit = pauli.min_qubit if min_qubit is None else min_qubit
        nqubit = pauli.nqubit if nqubit is None else nqubit

        statevector = self.run_statevector(
            circuit=circuit,
            statevector=statevector,
            min_qubit=min_qubit,
            nqubit=nqubit,
            dtype=dtype,
            **kwargs)

        pauli_expectation = PauliExpectation.zero()
        for string in pauli.keys():

            pauli2 = Pauli.zero()
            pauli2[string] = 1.0 
            statevector2 = self.run_pauli_sigma(
                pauli=pauli2,
                statevector=statevector,
                dtype=dtype,
                min_qubit=min_qubit,
                nqubit=nqubit,
                )
            scal = np.sum(statevector.conj() * statevector2)
            pauli_expectation[string] = scal

        return pauli_expectation

    def run_pauli_expectation_value_ideal(
        self,
        circuit,
        pauli,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        statevector = self.run_statevector(
            circuit=circuit,
            statevector=statevector,
            min_qubit=min_qubit,
            nqubit=nqubit,
            dtype=dtype,
            **kwargs)

        statevector2 = self.run_pauli_sigma(
            pauli=pauli,
            statevector=statevector,
            min_qubit=min_qubit,
            nqubit=nqubit,
            dtype=dtype,
            **kwargs)

        return np.sum(statevector.conj() * statevector2)

    # => Measurement-based Pauli expectations <= #

    # TODO: As always, there remains much work to be done in the conceptual,
    # pragmatical, and syntactical elements of this functionality

    def run_pauli_expectation_measurement(
        self,
        circuit,
        pauli,
        nmeasurement,
        statevector=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        **kwargs):

        min_qubit = pauli.min_qubit if min_qubit is None else min_qubit
        nqubit = pauli.nqubit if nqubit is None else nqubit

        # Determine commuting group
        groups = self.linear_commuting_group(
            pauli,
            min_qubit=min_qubit,
            nqubit=nqubit,
            )
        # Else exception will be raised if unknown commuting group
        # TODO: Optimally cover all commuting groups

        # Modified circuits for basis transformations
        circuits = [self.circuit_in_basis(
            circuit, 
            basis,
            min_qubit=min_qubit,
            nqubit=nqubit,
            ) 
            for basis in groups.keys()]
    
        # Measurements in commuting group (quantum heavy)
        probabilities = [self.run_measurement(
            circuit=circuit,
            nmeasurement=nmeasurement,
            statevector=statevector,
            min_qubit=min_qubit,
            nqubit=nqubit,
            dtype=dtype,
            **kwargs) for circuit in circuits]

        # Convert to counts
        results = [_.to_count_histogram() for _ in probabilities]

        # Counts for pauli strings
        counts = { _ : 0 for _ in pauli.keys() }
        ns = { _ : 0 for _ in pauli.keys() }
        for group, result in zip(groups.keys(), results):
            strings = groups[group]
            for string in strings:
                qubits = string.qubits
                ns[string] += nmeasurement
                for ket, count in result.items():
                    parity = sum((ket & (1 << (nqubit - 1 - (_ - min_qubit)))) >> (nqubit - 1 - (_ - min_qubit)) for _ in qubits) % 2
                    counts[string] += (-count) if parity else (+count)

        # Pauli density matrix values
        pauli_expectation = PauliExpectation(collections.OrderedDict([
            (_, counts[_] / max(ns[_], 1)) for _ in pauli.keys()]))
        if PauliString.I() in pauli_expectation:
            pauli_expectation[PauliString.I()] = 1.0 
        return pauli_expectation

    @staticmethod
    def linear_commuting_group(
        pauli,
        min_qubit=None,
        nqubit=None,
        ):

        min_qubit = pauli.min_qubit if min_qubit is None else min_qubit
        nqubit = pauli.nqubit if nqubit is None else nqubit

        keys = sortedcontainers.SortedSet()
        for string in pauli.keys():
            for qubit, char in string:
                keys.add(char)
        
        groups = collections.OrderedDict()
        for keyA in keys:
            for keyB in keys:
                groups[((keyA + keyB)*nqubit)[:nqubit]] = []

        for string in pauli.keys():

            # Do not do the identity operator
            if string.order == 0: continue

            # Add to all valid commuting groups
            found = False
            for group, strings2 in groups.items():
                valid = True
                for qubit, char in string:
                    if group[qubit - min_qubit] != char:
                        valid = False
                        break
                if not valid: continue
                strings2.append(string)
                found = True
            if not found: raise RuntimeError('Invalid string - not in linear commuting group: %s' % str(string))

        return groups

    @staticmethod
    def circuit_in_basis(
        circuit,
        basis,
        min_qubit=None,
        nqubit=None,
        ):

        min_qubit = circuit.min_qubit if min_qubit is None else min_qubit
        nqubit = circuit.nqubit if nqubit is None else nqubit

        if len(basis) != nqubit: raise RuntimeError('len(basis) != nqubit')

        basis_circuit = Circuit()
        for A, char in enumerate(basis): 
            qubit = A - min_qubit
            if char == 'X': basis_circuit.H(qubit)
            elif char == 'Y': basis_circuit.Rx2(qubit)
            elif char == 'Z': continue # Computational basis
            else: raise RuntimeError('Unknown basis: %s' % char)
    
        return Circuit.join_in_time([circuit, basis_circuit])
            
    # => Subset Hamiltonian Utilities <= #

    def run_pauli_matrix_subset(
        self,
        pauli,
        kets,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        ):

        min_qubit = pauli.min_qubit if min_qubit is None else min_qubit
        nqubit = pauli.nqubit if nqubit is None else nqubit

        if len(set(kets)) != len(kets): 
            raise RuntimeError('Kets are not unique')

        kets2 = { ket : index for index, ket in enumerate(kets) }

        H = np.zeros((len(kets),)*2, dtype=np.complex128)
        for ket_index, ket in enumerate(kets):
            for string, value in pauli.items():
                value = value + 0.j # Make sure value is complex
                bra = ket
                for qubit2, char in string:
                    qubit = qubit2 - min_qubit
                    if char == 'Z':
                        value *= -1.0 if (ket & (1 << (nqubit - 1 - qubit))) else 1.0
                    elif char == 'X':
                        bra ^= (1 << (nqubit - 1 - qubit))
                    elif char == 'Y':
                        value *= -1.j if (ket & (1 << (nqubit - 1 - qubit))) else 1.j
                        bra ^= (1 << (nqubit - 1 - qubit))
                bra_index = kets2.get(bra, None)
                if bra_index is None: continue
                H[bra_index, ket_index] += value
        
        return np.array(H, dtype=dtype)
                    
                 
        

        
        
    



import sortedcontainers

class IndexAllocator(object):

    """ IndexAllocator provides a utility object to track index
        allocation/deallocation, e.g., for reuse of ancilla indices.

        Example use:

        >>> allocator = IndexAllocator()
        >>> index0 = allocator.allocate()
        >>> index1 = allocator.allocate()
        >>> index2 = allocator.allocate()
        >>> print(allocator.next_index)
        >>> allocator.deallocate(index1)
        >>> print(allocator.next_index)
        
        For convenience, one can initialize a IndexAllocator with a number of
        pre-allocated indices via the build method:

        >>> allocator = IndexAllocator.build(10)
        >>> print(allocator.next_index)

        Negative indices are permitted throughout - next_index is always
        greater than min_index:

        >>> allocator = IndexAllocator()
        >>> allocator.allocate(-5)
        >>> print(allocator.next_index)

        Attributes:
            indices (SortedSet of int) - currently allocated indices.
    """
    def __init__(
        self,
        ):

        """ Empty IndexAllocator initializer. """

        self.indices = sortedcontainers.SortedSet()

    @staticmethod
    def build(
        nindex_reserved=0,
        ):

        """ Build a IndexAllocator with a number of pre-reserved indices.

        Params:
            nindex_reserved (int) - number of pre-reserved indices
        Returns:
            (IndexAllocator) - allocator with [0, nindex_reserved) indices
                already allocated.
        """

        allocator = IndexAllocator()
        for index in range(nindex_reserved):
            allocator.allocate()
        return allocator

    @property
    def min_index(self):
        """ (int) The minimum occupied index (or 0 if no occupied indices) """
        return self.indices[0] if len(self.indices) else 0
    
    @property
    def max_index(self):
        """ (int) The maximum occupied index (or -1 if no occupied indices) """
        return self.indices[-1] if len(self.indices) else -1

    @property
    def nindex(self):
        """ (int) The total number of indices (including empty indices). """
        return self.indices[-1] - self.indices[0] + 1 if len(self.indices) else 0

    @property
    def nindex_sparse(self):
        """ (int) The total number of occupied indices (excluding empty indices). """
        return len(self.indices)

    @property
    def next_index(self):
        """ (int) The next open index that remains unoccupied. """
        if self.nindex == 0: 
            return 0
        elif self.nindex == self.nindex_sparse: 
            return self.max_index + 1
        else: 
            for index in range(self.min_index, self.max_index):
                if index not in self.indices:
                    return index

    def allocate(self, index=None):
        """ Allocate an index.

        Params:
            index (int or None) - index to allocate. If None, allocates
                self.next_index. If index is already allocated, a RuntimeError
                will be raised.
        Return:
            (int) - allocated index, useful in case input index is None
                and self.next_index is allocated.
        """

        if index is None: index = self.next_index
        if index in self.indices: raise RuntimeError('Already allocated: %d' % index)
        self.indices.add(index)
        return index

    def deallocate(self, index):
        """ Deallocate an index.
        
        Params:
            index (int) - index to deallocate. If index is not allocated,
                a RuntimeError will be raised. 
        """
        
        if index not in self.indices: raise RuntimeError('Not allocated: %d' % index)
        self.indices.remove(index)

class NegativeIndexAllocator(IndexAllocator):

    """ NegativeIndexAllocator functions like IndexAllocator, except that it
        starts allocating at -1 and proceeds in steps of -1.
    """
    
    @property
    def next_index(self):
        """ (int) The next open index that remains unoccupied. """
        if self.nindex == 0: 
            return -1
        elif self.nindex == self.nindex_sparse: 
            return self.min_index - 1
        else: 
            for index in range(self.max_index, self.min_index, -1):
                if index not in self.indices:
                    return index



import sortedcontainers
from .circuit import ControlledGate


class Transpiler(object):

    """ class Transpiler is a collection of static utility functions that
        transform Circuit -> Circuit to exchange certain gate libraries. There
        are many potential use cases for these functions:

            - Flip control polarity on ControlledGates so that only positive
              controls are used.
            - Exchange ControlledGates for known Gates.
            - Remove redundant pairs of gates such as X * X or H * H.

        Many of these operations have preconditions on the required form of the
        input Circuit, such as:
            - Non-composite Circuit: the Circuit must have no CompositeGates
              (these drastically increase the complexity of transpilation
              operations). One should explode the Circuit before performing
              transpiler operations with this form.

        Many of these operations have known sequential application properties,
        such as:
            - Idempotence: op(op(circuit)) = op(circuit)
            - Nilpotence: op**N(circuit) = op**(N-1)(circuit) for some N

        These operations all return modified copies of the underlying circuits.
    """

    @staticmethod
    def polarize_controls(
        circuit,
        ):

        """ Add necessary X corrector gates to polarize all ControlledGate
            controls to True.

        Preconditions: Non-composite Circuit
        Transformation class: Idempotent
        """

        if circuit.is_composite: raise RuntimeError('circuit is composite')
        
        # Which starting times need X gates?
        expanders = sortedcontainers.SortedSet()
        for key, gate in circuit.gates.items():
            times, qubits = key
            if isinstance(gate, ControlledGate) and not all(gate.controls):
                expanders.add(times[0])
    
        # Map of old time -> new time with room for X corrector gates
        time_map = []
        counter = circuit.min_time
        for time in circuit.times:
            if time in expanders:
                time_map.append(counter+1)
                counter += 3
            else:
                time_map.append(counter)
                counter += 1

        # Circuit with room for X corrector gates
        circuit2 = circuit.slice(
            times=circuit.times, 
            times_to=time_map, 
            )

        # Perform the control polarization and mark X correctors
        X_gate_positions = []
        for key, gate in circuit2.gates.items():
            times, qubits = key
            if not isinstance(gate, ControlledGate) or all(gate.controls):
                continue
            for index, control in enumerate(gate.controls):
                if control: continue 
                # Mark X correctors
                X_gate_positions.append((times[0]-1, qubits[index]))
                X_gate_positions.append((times[0]+1, qubits[index]))
            # Polarize the controls
            gate.controls = [True]*gate.ncontrol

        # Add the X correctors
        for time, qubit in X_gate_positions:
            circuit2.X(qubit=qubit, times=time)

        return circuit2




import numpy as np
from .circuit import Circuit

class VariationalObservable(object):

    def __init__(
        self,
        backend,
        reference_weights,
        reference_circuits,
        entangler_circuit,
        entangler_parameter_group,
        pauli,
        nmeasurement=None,
        min_qubit=None,
        nqubit=None,
        dtype=np.complex128,
        ):

        self.backend = backend
        self.reference_weights = reference_weights
        self.reference_circuits = reference_circuits
        self.entangler_circuit = entangler_circuit
        self.entangler_parameter_group = entangler_parameter_group
        self.pauli = pauli
        self.nmeasurement = nmeasurement
        self.min_qubit = min_qubit
        self.nqubit = nqubit
        self.dtype = dtype
    
    @property
    def nparameter(self):
        return self.entangler_parameter_group.nparameter

    def run_observable(
        self,
        parameters,
        ):

        raw_parameters = self.entangler_parameter_group.compute_raw(parameters)
        entangler_circuit = self.entangler_circuit.copy()
        entangler_circuit.set_parameter_values(raw_parameters)

        E = 0.0
        for reference_weight, reference_circuit in zip(self.reference_weights, self.reference_circuits):
            circuit = Circuit.join_in_time([reference_circuit, entangler_circuit])
            E += reference_weight * self.backend.run_pauli_expectation_value(
                circuit=circuit,
                pauli=self.pauli,
                nmeasurement=self.nmeasurement,
                min_qubit=self.min_qubit,
                nqubit=self.nqubit,
                dtype=self.dtype,
                ).real

        return E
                
    def run_observable_gradient(
        self,
        parameters,
        ):

        raw_parameters = self.entangler_parameter_group.compute_raw(parameters)
        entangler_circuit = self.entangler_circuit.copy()
        entangler_circuit.set_parameter_values(raw_parameters)

        G = np.zeros((len(raw_parameters),), dtype=np.float64)
        for reference_weight, reference_circuit in zip(self.reference_weights, self.reference_circuits):
            circuit = Circuit.join_in_time([reference_circuit, entangler_circuit])
            G += reference_weight * self.backend.run_pauli_expectation_value_gradient(
                circuit=circuit,
                pauli=self.pauli,
                nmeasurement=self.nmeasurement,
                min_qubit=self.min_qubit,
                nqubit=self.nqubit,
                dtype=self.dtype,
                parameter_indices=list(range(reference_circuit.nparameter, reference_circuit.nparameter + entangler_circuit.nparameter)),
                ).real

        return self.entangler_parameter_group.compute_chain_rule1(parameters, G)




import numpy as np
from .measurement import ProbabilityHistogram
    
class Algebra(object):

    @staticmethod
    def apply_operator_1(
        statevector1,
        statevector2,
        operator,
        A,
        ):

        """ Apply a 1-body operator to statevector1 at qubit A, yielding statevector2.

        The formal operation performed is,

            statevector1_LiR = \\sum_{j} operator_ij statevector2_LjR

        Here L are the indices of all of the qubits to the left of A (<A), and
        R are the indices of all of the qubits to the right of A (>A).

        This function requires the user to supply both the initial state in
        statevector1 and an array statevector2 to place the result into. This allows this
        function to apply the gate without any new allocations or scratch arrays.

        Params:
            statevector1 (np.ndarray of shape (2**self.N,) and a complex dtype)
                - the initial statevector. operatornaffected by the operation
            statevector2 (np.ndarray of shape (2**self.N,) and a complex dtype)
                - an array to write the new statevector into. Overwritten by
                the operation.
            operator (np.ndarray of shape (2,2) and a complex dtype) - the matrix
                representation of the 1-body gate.
            A (int) - the qubit index to apply the gate at.
        Result:
            the data of statevector2 is overwritten with the result of the operation.
        Returns:
            reference to statevector2, for chaining
        """

        N = (statevector1.shape[0]&-statevector1.shape[0]).bit_length()-1
        if A >= N: raise RuntimeError('A >= N')
        if operator.shape != (2,2): raise RuntimeError('1-body gate must be (2,2)')
        if statevector1.shape != (2**N,): raise RuntimeError('statevector1 should be (%d,) shape, is %r shape' % (2**N, statevector1.shape))
        if statevector2.shape != (2**N,): raise RuntimeError('statevector2 should be (%d,) shape, is %r shape' % (2**N, statevector2.shape))

        L = 2**(A)     # Left hangover
        R = 2**(N-A-1) # Right hangover
        statevector1v = statevector1.view() 
        statevector2v = statevector2.view()
        statevector1v.shape = (L,2,R)
        statevector2v.shape = (L,2,R)
        np.einsum('LjR,ij->LiR', statevector1v, operator, out=statevector2v)

        return statevector2, statevector1

    @staticmethod
    def apply_operator_2(
        statevector1,
        statevector2,
        operator,
        A,
        B,
        ):

        """ Apply a 2-body operator to statevector1 at qubits A and B, yielding statevector2.

        The formal operation performed is (for the case that A < B),

            statevector1_LiMjR = \\sum_{lk} operator_ijkl statevector2_LiMjR

        Here L are the indices of all of the qubits to the left of A (<A), M M
        are the indices of all of the qubits to the right of A (>A) and left of
        B (<B), and R are the indices of all of the qubits to the right of B
        (>B). If A > B, permutations of A and B and the gate matrix operator are
        performed to ensure that the gate is applied correctly.

        This function requires the user to supply both the initial state in
        statevector1 and an array statevector2 to place the result into. This allows this
        function to apply the gate without any new allocations or scratch arrays.

        Params:
            statevector1 (np.ndarray of shape (2**self.N,) and a complex dtype)
                - the initial statevector. operatornaffected by the operation
            statevector2 (np.ndarray of shape (2**self.N,) and a complex dtype)
                - an array to write the new statevector into. Overwritten by
                the operation.
            operator (np.ndarray of shape (4,4) and a complex dtype) - the matrix
                representation of the 2-body gate. This should be packed to
                operate on the product state |A> otimes |B>, as usual.
            A (int) - the first qubit index to apply the gate at.
            B (int) - the second qubit index to apply the gate at.
        Result:
            the data of statevector2 is overwritten with the result of the operation.
        Returns:
            reference to statevector2, for chaining
        """

        N = (statevector1.shape[0]&-statevector1.shape[0]).bit_length()-1
        if A >= N: raise RuntimeError('A >= N')
        if B >= N: raise RuntimeError('B >= N')
        if A == B: raise RuntimeError('A == B')
        if operator.shape != (4,4): raise RuntimeError('2-body gate must be (4,4)')
        if statevector1.shape != (2**N,): raise RuntimeError('statevector1 should be (%d,) shape, is %r shape' % (2**N, statevector1.shape))
        if statevector2.shape != (2**N,): raise RuntimeError('statevector2 should be (%d,) shape, is %r shape' % (2**N, statevector2.shape))

        operator2 = np.reshape(operator, (2,2,2,2))
        if A > B:
            A2, B2 = B, A
            operator2 = np.einsum('ijkl->jilk', operator2)
        else:
            A2, B2 = A, B

        L = 2**(A2)      # Left hangover
        M = 2**(B2-A2-1) # Middle hangover
        R = 2**(N-B2-1)  # Right hangover
        statevector1v = statevector1.view() 
        statevector2v = statevector2.view()
        statevector1v.shape = (L,2,M,2,R)
        statevector2v.shape = (L,2,M,2,R)
        np.einsum('LkMlR,ijkl->LiMjR', statevector1v, operator2, out=statevector2v)

        return statevector2, statevector1

    @staticmethod
    def apply_operator_3(
        statevector1,
        statevector2,
        operator,
        A,
        B,
        C,
        ):

        """ Apply a 3-body operator to statevector1 at qubits A, B, and C, yielding statevector2.

        This function requires the user to supply both the initial state in
        statevector1 and an array statevector2 to place the result into. This allows this
        function to apply the gate without any new allocations or scratch arrays.

        Params:
            statevector1 (np.ndarray of shape (2**self.N,) and a complex dtype)
                - the initial statevector. operatornaffected by the operation
            statevector2 (np.ndarray of shape (2**self.N,) and a complex dtype)
                - an array to write the new statevector into. Overwritten by
                the operation.
            operator (np.ndarray of shape (8,8) and a complex dtype) - the matrix
                representation of the 3-body gate. This should be packed to
                operate on the product state |A> otimes |B> otimes |C>, as
                usual.
            A (int) - the first qubit index to apply the gate at.
            B (int) - the second qubit index to apply the gate at.
            C (int) - the third qubit index to apply the gate at.
        Result:
            the data of statevector2 is overwritten with the result of the operation.
        Returns:
            reference to statevector2, for chaining
        """

        N = (statevector1.shape[0]&-statevector1.shape[0]).bit_length()-1
        if A >= N: raise RuntimeError('A >= N')
        if B >= N: raise RuntimeError('B >= N')
        if C >= N: raise RuntimeError('C >= N')
        if A == B: raise RuntimeError('A == B')
        if A == C: raise RuntimeError('A == C')
        if B == C: raise RuntimeError('B == C')
        if operator.shape != (8,8): raise RuntimeError('3-body gate must be (8,8)')
        if statevector1.shape != (2**N,): raise RuntimeError('statevector1 should be (%d,) shape, is %r shape' % (2**N, statevector1.shape))
        if statevector2.shape != (2**N,): raise RuntimeError('statevector2 should be (%d,) shape, is %r shape' % (2**N, statevector2.shape))

        A2, B2, C2 = sorted((A, B, C))

        operator2 = np.reshape(operator, (2,2,2,2,2,2))

        bra_indices = 'ijk'
        ket_indices = 'lmn'
        bra_indices2 = ''.join([bra_indices[(A, B, C).index(_)] for _ in (A2, B2, C2)])
        ket_indices2 = ''.join([ket_indices[(A, B, C).index(_)] for _ in (A2, B2, C2)])
        
        operator2 = np.einsum('%s%s->%s%s' % (bra_indices, ket_indices, bra_indices2, ket_indices2), operator2)

        L = 2**(A2)      # Left hangover
        M = 2**(B2-A2-1) # Middle1 hangover
        P = 2**(C2-B2-1) # Middle2 hangover
        R = 2**(N-C2-1)  # Right hangover
        statevector1v = statevector1.view() 
        statevector2v = statevector2.view()
        statevector1v.shape = (L,2,M,2,P,2,R)
        statevector2v.shape = (L,2,M,2,P,2,R)
        np.einsum('LlMmPnR,ijklmn->LiMjPkR', statevector1v, operator2, out=statevector2v)

        return statevector2, statevector1
    
    @staticmethod
    def apply_operator_n(
        statevector1,
        statevector2,
        operator,
        qubits,
        ):

        N = (statevector1.shape[0]&-statevector1.shape[0]).bit_length()-1
        if any(_ >= N for _ in qubits): raise RuntimeError('qubits >= N')
        if len(set(qubits)) != len(qubits): raise RuntimeError('duplicate entry in qubits')
        if statevector1.shape != (2**N,): raise RuntimeError('statevector1 should be (%d,) shape, is %r shape' % (2**N, statevector1.shape))
        if statevector2.shape != (2**N,): raise RuntimeError('statevector2 should be (%d,) shape, is %r shape' % (2**N, statevector2.shape))
           
        # hangover
        qubits2 = tuple(sorted(qubits))
        hangovers = (2**qubits2[0],) + tuple(2**(qubits2[A+1]-qubits2[A]-1) for A in range(len(qubits2)-1)) + (2**(N-qubits2[-1]-1),)
        shape = []
        for hangover in hangovers[:-1]:
            shape.append(hangover)
            shape.append(2)
        shape.append(hangovers[-1])
        shape = tuple(shape)
        
        statevector1v = statevector1.view() 
        statevector2v = statevector2.view()
        statevector1v.shape = shape
        statevector2v.shape = shape
        
        # symbol stock 
        hangover_stock = 'ABCDEFGHIJKLMNOPQRSToperatorVWXYZ'
        bra_stock = 'abcdefghijklm'
        ket_stock = 'nopqrstuvwxyz'
    
        M = len(qubits)
        if M > 13: raise RuntimeError('Technical limit: cannot run N > 13')
        
        # einsum form for ordering gate
        bra_str = bra_stock[:M]
        ket_str = ket_stock[:M]
        bra_str2 = ''.join([bra_str[qubits.index(_)] for _ in qubits2])
        ket_str2 = ''.join([ket_str[qubits.index(_)] for _ in qubits2])
        operator_str = ket_stock[:M] + bra_stock[:M]
        
        shape_operator = tuple(2 for _ in range(2*M))
        operator2 = np.reshape(operator, shape_operator)
        operator2 = np.einsum('%s%s->%s%s' % (bra_str, ket_str, bra_str2, ket_str2), operator2)
        
        # einsum form for applying gate
        statevector1_str = ''
        statevector2_str = ''
        for A in range(M):
            statevector1_str += hangover_stock[A]
            statevector1_str += bra_stock[A]
            statevector2_str += hangover_stock[A]
            statevector2_str += ket_stock[A]
        statevector1_str += hangover_stock[M]
        statevector2_str += hangover_stock[M]
        
        np.einsum('%s,%s->%s' % (statevector1_str, operator_str, statevector2_str), statevector1v, operator2, out=statevector2v)
        statevector2v = np.reshape(statevector2v, (-1,))
        
        return statevector2, statevector1

    @staticmethod
    def sample_histogram_from_probabilities(
        probabilities,
        nmeasurement=None,
        cutoff=1.0E-12,
        ):

        """ Randomly sample binary string measurements from a set of
            probabilities of size 2**N.

        Params:
            statevector (np.ndarray of shape (2**N,)) - the statevector to
                sample.
            nmeasurement (int or None) - number of measurements to sample. If
                None, infinite sampling is assumed and the probabilities are
                directly returned in ProbabilityHistogram format.
            cutoff (float) - probability below which to ignore contributions.
        Returns:
            (ProbabilityHistogram) - a ProbabilityHistogram object containing
                the results of randomly sampled projective measurements.
        """

        N = (probabilities.shape[0]&-probabilities.shape[0]).bit_length()-1
    
        # Directly return probabilities if nmeasurement is None (infinite sampling)
        if nmeasurement is None:
            return ProbabilityHistogram(
                nqubit=N,
                histogram={ k : v for k, v in enumerate(probabilities) if v > cutoff},
                nmeasurement=nmeasurement,
                ) 
        # Otherwise, perform random sampling
        if not isinstance(nmeasurement, int):
            raise RuntimeError('nmeasurement must be int: %s' % nmeasurement)
        I = list(np.searchsorted(np.cumsum(probabilities), np.random.rand(nmeasurement)))
        return ProbabilityHistogram(
            nqubit=N,
            histogram={ int(k) : I.count(k) / nmeasurement for k in list(sorted(set(I))) },
            nmeasurement=nmeasurement,
            ) 



import quasar
import unittest
import numpy as np

class TestMatrix(unittest.TestCase):

    def helper_test_matrices(
        self,
        quasar_matrices,
        reference_matrices,
        ):

        for key, reference_matrix in reference_matrices.items():
            quasar_matrix = quasar_matrices[key]
            self.assertIsInstance(quasar_matrix, np.ndarray)
            self.assertEqual(quasar_matrix.dtype, np.complex128)
            self.assertEqual(quasar_matrix.shape, reference_matrix.shape)
            self.assertTrue(np.max(np.abs(quasar_matrix - reference_matrix)) < 1.0E-14)

    def test_one_qubit_raw(self):

        quasar_matrices = {
            'I'  : quasar.Matrix.I,
            'X'  : quasar.Matrix.X,
            'Y'  : quasar.Matrix.Y,
            'Z'  : quasar.Matrix.Z,
            'S'  : quasar.Matrix.S,
            'ST' : quasar.Matrix.ST,
            'T'  : quasar.Matrix.T,
            'TT' : quasar.Matrix.TT,
            'H'  : quasar.Matrix.H,
        }

        reference_matrices = {
            'I'  : np.array([[1.0, 0.0], [0.0, 1.0]], dtype=np.complex128),
            'X'  : np.array([[0.0, 1.0], [1.0, 0.0]], dtype=np.complex128),
            'Y'  : np.array([[0.0, -1.0j], [+1.0j, 0.0]], dtype=np.complex128),
            'Z'  : np.array([[1.0, 0.0], [0.0, -1.0]], dtype=np.complex128),
            'S'  : np.array([[1.0, 0.0], [0.0, +1.0j]], dtype=np.complex128),
            'ST' : np.array([[1.0, 0.0], [0.0, -1.0j]], dtype=np.complex128),
            'T'  : np.array([[1.0, 0.0], [0.0, np.cos(np.pi / 4.0) + 1.j * np.sin(np.pi / 4.0)]], dtype=np.complex128),
            'TT' : np.array([[1.0, 0.0], [0.0, np.cos(np.pi / 4.0) - 1.j * np.sin(np.pi / 4.0)]], dtype=np.complex128),
            'H'  : np.array([[1.0, 1.0], [1.0, -1.0]], dtype=np.complex128) / np.sqrt(2.0)
        }

        self.helper_test_matrices(quasar_matrices, reference_matrices)

            
        
 



import collections
import numpy as np

class Ket(int):
    
    def __getitem__(
        self,
        key,
        ):

        return (self & (1 << key)) >> key

    def 

class Ket(str):

    def __new__(
        self,
        string,
        ):

        if not isinstance(string, str): raise RuntimeError('string must be str')
        if not all(_ == '0' or _ == '1' for _ in string): raise RuntimeError('Invalid Ket: %s' % string)

        return str.__new__(Ket, string)

    def __getitem__(
        self,
        key,
        ):

        return int(super(Ket, self).__getitem__(key))

    @property
    def nqubit(self):
        return len(self)

    @staticmethod
    def from_int(value, nqubit):
        if value >= (1 << nqubit): raise RuntimeError('value >= 2**nqubit: value=%s, nqubit=%s' % (value, nqubit))
        if value<0: raise RuntimeError('value must be a positive int')
        if not isinstance(value, int): raise RuntimeError('value must be int')
        start = bin(value)[2:]
        padded = ('0' * (nqubit - len(start))) + start
        return Ket(padded)

    @staticmethod
    def from_int_reverse(value, nqubit):
        if value >= (1 << nqubit): raise RuntimeError('value >= 2**nqubit: value=%s, nqubit=%s' % (value, nqubit))
        if value<0: raise RuntimeError('value must be a positive int')
        if not isinstance(value, int): raise RuntimeError('value must be int')
        start = bin(value)[2:]
        padded = ('0' * (nqubit - len(start))) + start
        return Ket(padded[::-1])

class Histogram(object):

    def items(self):
        return self.histogram.items()

    def keys(self):
        return self.histogram.keys()

    def values(self):
        return self.histogram.values()

    def __contains__(
        self,
        key,
        ):

        key = Ket(key)
        return self.histogram.__contains__(key)

    def __getitem__(
        self,
        key,
        ):

        key = Ket(key)
        return self.histogram.__getitem__(key)

    def get(
        self,
        key,
        default=None,
        ):

        key = Ket(key)
        return self.histogram.get(key, default)

    @property
    def nqubit(self):
        return next(iter(self.histogram.keys())).nqubit if len(self.histogram) else None

class ProbabilityHistogram(Histogram):

    def __init__(
        self,
        histogram={},
        nmeasurement=None,
        ):

        self.histogram = histogram.copy()
        self.nmeasurement = nmeasurement

        for k, v in self.histogram.items():
            if not isinstance(k, Ket): raise RuntimeError('Key must be Ket: %s' % k) 
            if not isinstance(v, float): raise RuntimeError('Value must be float: %s' % v) 

        if len(self.histogram) == 0: return
        if not all(_.nqubit == self.nqubit for _ in self.histogram.keys()): raise RuntimeError('All keys must have same nqubit')

    def __str__(self):
        s = ''
        s += 'nmeasurement : %r\n' % (self.nmeasurement) 
        for k in sorted(self.histogram.keys()):
            s += '|%s> : %8.6f\n' % (k, self.histogram[k])
        return s

    def to_count_histogram(self):
        if self.nmeasurement is None:
            raise RuntimeError('Cannot convert to count histogram: nmeasurement is None (infinite sampling)')
        return CountHistogram(
            histogram={ k : int(round(v * self.nmeasurement)) for k, v in self.items() },
            nmeasurement=self.nmeasurement,
            )

class CountHistogram(Histogram):

    def __init__(
        self,
        histogram={},
        nmeasurement=None,
        ):

        self.histogram = histogram.copy()
        self.nmeasurement = nmeasurement

        if self.nmeasurement is None: 
            raise RuntimeError('nmeasurement cannot be None in CountHistogram')

        for k, v in self.histogram.items():
            if not isinstance(k, Ket): raise RuntimeError('Key must be Ket: %s' % k) 
            if not isinstance(v, int): raise RuntimeError('Value must be int: %s' % v) 

        if len(self.histogram) == 0: return
        if not all(_.nqubit == self.nqubit for _ in self.histogram.keys()): raise RuntimeError('All keys must have same nqubit')
        if sum(self.histogram.values()) != self.nmeasurement: raise RuntimeError('Values do not sum to nmeasurement')

    def __str__(self):
        maxval = max(_ for _ in self.histogram.values())
        maxlen = max(1, int(np.floor(np.log10(maxval))) + 1)        
        s = ''
        s += 'nmeasurement : %r\n' % (self.nmeasurement) 
        for k in sorted(self.histogram.keys()):
            s += '|%s> : %*d\n' % (k, maxlen, self.histogram[k])
        return s

    def to_probability_histogram(self):
        if self.nmeasurement is None:
            raise RuntimeError('nmeasurement cannot be None in CountHistogram')
        return ProbabilityHistogram(
            histogram={ k : v / self.nmeasurement for k, v in self.items() },
            nmeasurement=self.nmeasurement,
            )



import quasar
import unittest
import sortedcontainers

class TestCircuit(unittest.TestCase):

    def test_empty_circuit(self):

        circuit = quasar.Circuit()
        
        self.assertEqual(circuit.ngate, 0)
        self.assertEqual(circuit.ngate1, 0)
        self.assertEqual(circuit.ngate2, 0)
        self.assertEqual(circuit.ngate3, 0)
        self.assertEqual(circuit.ngate4, 0)
        self.assertEqual(circuit.ngate_nqubit(0), 0)
        self.assertEqual(circuit.ngate_nqubit(1), 0)
        self.assertEqual(circuit.ngate_nqubit(2), 0)
        self.assertEqual(circuit.ngate_nqubit(3), 0)
        self.assertEqual(circuit.ngate_nqubit(4), 0)
        self.assertEqual(circuit.max_gate_nqubit, 0)
        self.assertEqual(circuit.max_gate_ntime, 0)
        self.assertEqual(circuit.min_time, 0)
        self.assertEqual(circuit.max_time, -1)
        self.assertEqual(circuit.ntime, 0)
        self.assertEqual(circuit.ntime_sparse, 0)
        self.assertEqual(circuit.min_qubit, 0)
        self.assertEqual(circuit.max_qubit, -1)
        self.assertEqual(circuit.nqubit, 0)
        self.assertEqual(circuit.nqubit_sparse, 0)
        self.assertFalse(circuit.is_controlled)
        self.assertFalse(circuit.is_composite)

        self.assertIsInstance(circuit, quasar.Circuit)
        self.assertIsInstance(circuit.gates, sortedcontainers.SortedDict)
        self.assertIsInstance(circuit.times, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.qubits, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.times_and_qubits, sortedcontainers.SortedSet)

        self.assertEqual(circuit.gates, sortedcontainers.SortedDict([]))
        self.assertEqual(circuit.times, sortedcontainers.SortedSet([]))
        self.assertEqual(circuit.qubits, sortedcontainers.SortedSet([]))
        self.assertEqual(circuit.times_and_qubits, sortedcontainers.SortedSet([]))

        self.assertIsInstance(str(circuit), str)

    def test_h_circuit(self): 
        
        circuit = quasar.Circuit().add_gate(quasar.Gate.H, 0, copy=False)
        
        self.assertEqual(circuit.ngate, 1)
        self.assertEqual(circuit.ngate1, 1)
        self.assertEqual(circuit.ngate2, 0)
        self.assertEqual(circuit.ngate3, 0)
        self.assertEqual(circuit.ngate4, 0)
        self.assertEqual(circuit.ngate_nqubit(0), 0)
        self.assertEqual(circuit.ngate_nqubit(1), 1)
        self.assertEqual(circuit.ngate_nqubit(2), 0)
        self.assertEqual(circuit.ngate_nqubit(3), 0)
        self.assertEqual(circuit.ngate_nqubit(4), 0)
        self.assertEqual(circuit.max_gate_nqubit, 1)
        self.assertEqual(circuit.max_gate_ntime, 1)
        self.assertEqual(circuit.min_time, 0)
        self.assertEqual(circuit.max_time, 0)
        self.assertEqual(circuit.ntime, 1)
        self.assertEqual(circuit.ntime_sparse, 1)
        self.assertEqual(circuit.min_qubit, 0)
        self.assertEqual(circuit.max_qubit, 0)
        self.assertEqual(circuit.nqubit, 1)
        self.assertEqual(circuit.nqubit_sparse, 1)
        self.assertFalse(circuit.is_controlled)
        self.assertFalse(circuit.is_composite)

        self.assertIsInstance(circuit, quasar.Circuit)
        self.assertIsInstance(circuit.gates, sortedcontainers.SortedDict)
        self.assertIsInstance(circuit.times, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.qubits, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.times_and_qubits, sortedcontainers.SortedSet)

        self.assertEqual(circuit.gates, sortedcontainers.SortedDict([(((0,), (0,)), quasar.Gate.H)]))
        self.assertEqual(circuit.times, sortedcontainers.SortedSet([0]))
        self.assertEqual(circuit.qubits, sortedcontainers.SortedSet([0]))
        self.assertEqual(circuit.times_and_qubits, sortedcontainers.SortedSet([(0, 0)]))

        self.assertIsInstance(str(circuit), str)

    def test_ghz_circuit(self): 
        
        circuit = quasar.Circuit()
        circuit.add_gate(quasar.Gate.H, 0, copy=False)
        circuit.add_gate(quasar.Gate.CX, (0,1), copy=False)
        circuit.add_gate(quasar.Gate.CX, qubits=(1,2), copy=False)
        
        self.assertEqual(circuit.ngate, 3)
        self.assertEqual(circuit.ngate1, 1)
        self.assertEqual(circuit.ngate2, 2)
        self.assertEqual(circuit.ngate3, 0)
        self.assertEqual(circuit.ngate4, 0)
        self.assertEqual(circuit.ngate_nqubit(0), 0)
        self.assertEqual(circuit.ngate_nqubit(1), 1)
        self.assertEqual(circuit.ngate_nqubit(2), 2)
        self.assertEqual(circuit.ngate_nqubit(3), 0)
        self.assertEqual(circuit.ngate_nqubit(4), 0)
        self.assertEqual(circuit.max_gate_nqubit, 2)
        self.assertEqual(circuit.max_gate_ntime, 1)
        self.assertEqual(circuit.min_time, 0)
        self.assertEqual(circuit.max_time, 2)
        self.assertEqual(circuit.ntime, 3)
        self.assertEqual(circuit.ntime_sparse, 3)
        self.assertEqual(circuit.min_qubit, 0)
        self.assertEqual(circuit.max_qubit, 2)
        self.assertEqual(circuit.nqubit, 3)
        self.assertEqual(circuit.nqubit_sparse, 3)
        self.assertFalse(circuit.is_controlled)
        self.assertFalse(circuit.is_composite)

        self.assertIsInstance(circuit, quasar.Circuit)
        self.assertIsInstance(circuit.gates, sortedcontainers.SortedDict)
        self.assertIsInstance(circuit.times, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.qubits, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.times_and_qubits, sortedcontainers.SortedSet)

        self.assertEqual(circuit.gates, sortedcontainers.SortedDict([
            (((0,), (0,)), quasar.Gate.H),
            (((1,), (0, 1)), quasar.Gate.CX),
            (((2,), (1, 2)), quasar.Gate.CX),
            ]))
        self.assertEqual(circuit.times, sortedcontainers.SortedSet([0, 1, 2]))
        self.assertEqual(circuit.qubits, sortedcontainers.SortedSet([0, 1, 2]))
        self.assertEqual(circuit.times_and_qubits, sortedcontainers.SortedSet([(0, 0), (1, 0), (1, 1), (2, 1), (2, 2)]))

        self.assertIsInstance(str(circuit), str)

    def test_sparse_circuit(self): 
        
        circuit = quasar.Circuit()
        circuit.add_gate(quasar.Gate.H, 0, copy=False)
        circuit.add_gate(quasar.Gate.CX, (0,1), times=(2,), copy=False)
        circuit.add_gate(quasar.Gate.CX, qubits=(3,4), times=(4,), copy=False)
        
        self.assertEqual(circuit.ngate, 3)
        self.assertEqual(circuit.ngate1, 1)
        self.assertEqual(circuit.ngate2, 2)
        self.assertEqual(circuit.ngate3, 0)
        self.assertEqual(circuit.ngate4, 0)
        self.assertEqual(circuit.ngate_nqubit(0), 0)
        self.assertEqual(circuit.ngate_nqubit(1), 1)
        self.assertEqual(circuit.ngate_nqubit(2), 2)
        self.assertEqual(circuit.ngate_nqubit(3), 0)
        self.assertEqual(circuit.ngate_nqubit(4), 0)
        self.assertEqual(circuit.max_gate_nqubit, 2)
        self.assertEqual(circuit.max_gate_ntime, 1)
        self.assertEqual(circuit.min_time, 0)
        self.assertEqual(circuit.max_time, 4)
        self.assertEqual(circuit.ntime, 5)
        self.assertEqual(circuit.ntime_sparse, 3)
        self.assertEqual(circuit.min_qubit, 0)
        self.assertEqual(circuit.max_qubit, 4)
        self.assertEqual(circuit.nqubit, 5)
        self.assertEqual(circuit.nqubit_sparse, 4)
        self.assertFalse(circuit.is_controlled)
        self.assertFalse(circuit.is_composite)

        self.assertIsInstance(circuit, quasar.Circuit)
        self.assertIsInstance(circuit.gates, sortedcontainers.SortedDict)
        self.assertIsInstance(circuit.times, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.qubits, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.times_and_qubits, sortedcontainers.SortedSet)

        self.assertEqual(circuit.gates, sortedcontainers.SortedDict([
            (((0,), (0,)), quasar.Gate.H),
            (((2,), (0, 1)), quasar.Gate.CX),
            (((4,), (3, 4)), quasar.Gate.CX),
            ]))
        self.assertEqual(circuit.times, sortedcontainers.SortedSet([0, 2, 4]))
        self.assertEqual(circuit.qubits, sortedcontainers.SortedSet([0, 1, 3, 4]))
        self.assertEqual(circuit.times_and_qubits, sortedcontainers.SortedSet([(0, 0), (2, 0), (2, 1), (4, 3), (4, 4)]))

        self.assertIsInstance(str(circuit), str)

    def test_shifted_sparse_circuit(self): 
        
        circuit = quasar.Circuit()
        circuit.add_gate(quasar.Gate.H, -2, times=-2, copy=False)
        circuit.add_gate(quasar.Gate.CX, (0,1), times=(2,), copy=False)
        circuit.add_gate(quasar.Gate.CX, qubits=(3,4), times=(4,), copy=False)
        
        self.assertEqual(circuit.ngate, 3)
        self.assertEqual(circuit.ngate1, 1)
        self.assertEqual(circuit.ngate2, 2)
        self.assertEqual(circuit.ngate3, 0)
        self.assertEqual(circuit.ngate4, 0)
        self.assertEqual(circuit.ngate_nqubit(0), 0)
        self.assertEqual(circuit.ngate_nqubit(1), 1)
        self.assertEqual(circuit.ngate_nqubit(2), 2)
        self.assertEqual(circuit.ngate_nqubit(3), 0)
        self.assertEqual(circuit.ngate_nqubit(4), 0)
        self.assertEqual(circuit.max_gate_nqubit, 2)
        self.assertEqual(circuit.max_gate_ntime, 1)
        self.assertEqual(circuit.min_time, -2)
        self.assertEqual(circuit.max_time, 4)
        self.assertEqual(circuit.ntime, 7)
        self.assertEqual(circuit.ntime_sparse, 3)
        self.assertEqual(circuit.min_qubit, -2)
        self.assertEqual(circuit.max_qubit, 4)
        self.assertEqual(circuit.nqubit, 7)
        self.assertEqual(circuit.nqubit_sparse, 5)
        self.assertFalse(circuit.is_controlled)
        self.assertFalse(circuit.is_composite)

        self.assertIsInstance(circuit, quasar.Circuit)
        self.assertIsInstance(circuit.gates, sortedcontainers.SortedDict)
        self.assertIsInstance(circuit.times, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.qubits, sortedcontainers.SortedSet)
        self.assertIsInstance(circuit.times_and_qubits, sortedcontainers.SortedSet)

        self.assertEqual(circuit.gates, sortedcontainers.SortedDict([
            (((-2,), (-2,)), quasar.Gate.H),
            (((2,), (0, 1)), quasar.Gate.CX),
            (((4,), (3, 4)), quasar.Gate.CX),
            ]))
        self.assertEqual(circuit.times, sortedcontainers.SortedSet([-2, 2, 4]))
        self.assertEqual(circuit.qubits, sortedcontainers.SortedSet([-2, 0, 1, 3, 4]))
        self.assertEqual(circuit.times_and_qubits, sortedcontainers.SortedSet([(-2, -2), (2, 0), (2, 1), (4, 3), (4, 4)]))

        self.assertIsInstance(str(circuit), str)

    def test_circuit_equivalence(self):
        
        # Reflexivity
        circuit1 = quasar.Circuit().H(0).CX(0, 1)
        self.assertTrue(quasar.Circuit.test_equivalence(circuit1, circuit1))

        # Reflexivity (different objects)
        circuit1 = quasar.Circuit().H(0).CX(0, 1)
        circuit2 = quasar.Circuit().H(0).CX(0, 1)
        self.assertTrue(quasar.Circuit.test_equivalence(circuit1, circuit2))

        # Reflexivity (different build order)
        circuit1 = quasar.Circuit().CX(0, 1, times=1).H(0, times=0)
        circuit2 = quasar.Circuit().H(0).CX(0, 1)
        self.assertTrue(quasar.Circuit.test_equivalence(circuit1, circuit2))

        # Same layout, different gates
        circuit1 = quasar.Circuit().Z(0).CX(0, 1)
        circuit2 = quasar.Circuit().H(0).CX(0, 1)
        self.assertFalse(quasar.Circuit.test_equivalence(circuit1, circuit2))

        # Same layout, different times
        circuit1 = quasar.Circuit().H(0).CX(0, 1)
        circuit2 = quasar.Circuit().H(0, time_start=-1).CX(1, 2)
        self.assertFalse(quasar.Circuit.test_equivalence(circuit1, circuit2))

        # Same layout, different qubits
        circuit1 = quasar.Circuit().H(0).CX(0, 1)
        circuit2 = quasar.Circuit().H(1).CX(1, 2)
        self.assertFalse(quasar.Circuit.test_equivalence(circuit1, circuit2))

if __name__ == '__main__':

    unittest.main()



"""
Build and execute quantum circuits on various backends.

Quasar is a package to simplify the building and execution of quantum computing
circuits for quantum computing in a reasonably vendor-agnostic way.
It provides backends for multiple providers, but also has a credible
classical simulator.  It is also the basis for circuit-model functionality
in QCWare's Forge platform (http://www.qcware.com)
"""
__version__ = '1.0.0'
from .circuit import Matrix, Gate, Circuit
from .circuit import ControlledGate, CompositeGate
from .pauli import PauliStarter, PauliOperator, PauliString, Pauli, PauliExpectation
from .pauli import PauliOperator
from .index_allocator import IndexAllocator, NegativeIndexAllocator
from .transpiler import Transpiler
from .backend import Backend
from .quasar_backend import QuasarSimulatorBackend
from .measurement import Histogram, ProbabilityHistogram, CountHistogram
from .format import format_statevector
from .davidson import Davidson, run_davidson
from .parameters import ParameterGroup, FixedParameterGroup, IdentityParameterGroup, LinearParameterGroup, CompositeParameterGroup
from .observable import VariationalObservable
from .optimizer import Optimizer, BFGSOptimizer
from .options import Options



import numpy as np

class Histogram(object):

    def items(self):
        return self.histogram.items()

    def keys(self):
        return self.histogram.keys()

    def values(self):
        return self.histogram.values()

    def __contains__(
        self,
        key,
        ):

        if isinstance(key, str): key = int(key, base=2)
        return self.histogram.__contains__(key)

    def __getitem__(
        self,
        key,
        ):

        if isinstance(key, str): key = int(key, base=2)
        return self.histogram.__getitem__(key)

    def get(
        self,
        key,
        default=None,
        ):

        if isinstance(key, str): key = int(key, base=2)
        return self.histogram.get(key, default)

class ProbabilityHistogram(Histogram):

    def __init__(
        self,
        nqubit,
        histogram={},
        nmeasurement=None,
        ):

        self.nqubit = nqubit
        self.histogram = histogram.copy()
        self.nmeasurement = nmeasurement

        if not isinstance(self.nqubit, int): 
            raise RuntimeError('nqubit must be int')
        if self.nmeasurement is not None and not isinstance(self.nmeasurement, (int)):
            raise RuntimeError('nmeasurement must be int or None')
        if not all(isinstance(key, int) for key in self.histogram.keys()):
            raise RuntimeError('Keys must be int')
        if not all(isinstance(value, float) for value in self.histogram.values()):
            raise RuntimeError('Value must be float')

    def __str__(self): 
    
        s = ''
        s += 'nqubit       : %r\n' % (self.nqubit)
        s += 'nmeasurement : %r\n' % (self.nmeasurement)
        for k in sorted(self.histogram.keys()):
            ket = bin(k)[2:]
            ket = '0' * (self.nqubit - len(ket)) + ket
            s += '|%s> : %8.6f\n' % (ket, self.histogram[k])
        return s

    def to_count_histogram(self):
        if self.nmeasurement is None:
            raise RuntimeError('Cannot convert to count histogram: nmeasurement is None (infinite sampling)')
        return CountHistogram(
            nqubit=self.nqubit,
            histogram={ k : int(round(v * self.nmeasurement)) for k, v in self.items() },
            nmeasurement=self.nmeasurement,
            )

    def subset(
        self,
        qubits,
        ):

        if len(set(qubits)) != len(qubits): raise RuntimeError('qubits are not unique')
        if any(_ < 0 or _ >= self.nqubit for _ in qubits): raise RuntimeError('qubits do not lie in [0, nqubit)')

        nqubit2 = len(qubits)
        histogram2 = {}
        for key, value in self.histogram.items():
            key2 = 0
            for qubit2, qubit in enumerate(qubits):
                key2 += ((key & (1 << (self.nqubit - 1 - qubit))) >> (self.nqubit - 1 - qubit)) << (nqubit2 - 1 - qubit2)
            histogram2[key2] = value + histogram2.get(key2, 0.0)

        return ProbabilityHistogram(
            nqubit=nqubit2,
            histogram=histogram2,
            nmeasurement=self.nmeasurement,
            )

class CountHistogram(Histogram):

    def __init__(
        self,
        nqubit,
        histogram={},
        nmeasurement=None,
        ):

        self.nqubit = nqubit
        self.histogram = histogram.copy()
        self.nmeasurement = nmeasurement

        if not isinstance(self.nqubit, int): 
            raise RuntimeError('nqubit must be int')
        if not isinstance(self.nmeasurement, (int)):
            raise RuntimeError('nmeasurement must be int')
        if not all(isinstance(key, int) for key in self.histogram.keys()):
            raise RuntimeError('Keys must be int')
        if not all(isinstance(value, int) for value in self.histogram.values()):
            raise RuntimeError('Value must be int')

    def __str__(self): 
    
        maxval = max(_ for _ in self.histogram.values())
        maxlen = max(1, int(np.floor(np.log10(maxval))) + 1)        
        s = ''
        s += 'nqubit       : %r\n' % (self.nqubit)
        s += 'nmeasurement : %r\n' % (self.nmeasurement)
        for k in sorted(self.histogram.keys()):
            ket = bin(k)[2:]
            ket = '0' * (self.nqubit - len(ket)) + ket
            s += '|%s> : %*d\n' % (ket, maxlen, self.histogram[k])
        return s

    def to_probability_histogram(self):
        if self.nmeasurement is None:
            raise RuntimeError('nmeasurement cannot be None in CountHistogram')
        return ProbabilityHistogram(
            nqubit=self.nqubit,
            histogram={ k : v / self.nmeasurement for k, v in self.items() },
            nmeasurement=self.nmeasurement,
            )

    def subset(
        self,
        qubits,
        ):

        if len(set(qubits)) != len(qubits): raise RuntimeError('qubits are not unique')
        if any(_ < 0 or _ >= self.nqubit for _ in qubits): raise RuntimeError('qubits do not lie in [0, nqubit)')

        nqubit2 = len(qubits)
        histogram2 = {}
        for key, value in self.histogram.items():
            key2 = 0
            for qubit2, qubit in enumerate(qubits):
                key2 += ((key & (1 << (self.nqubit - 1 - qubit))) >> (self.nqubit - 1 - qubit)) << (nqubit2 - 1 - qubit2)
            histogram2[key2] = value + histogram2.get(key2, 0)

        return CountHistogram(
            nqubit=nqubit2,
            histogram=histogram2,
            nmeasurement=self.nmeasurement,
            )



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['quasar']

package_data = \
{'': ['*']}

install_requires = \
['numpy >= 1.18.4', 'scipy >= 1.3.3', 'sortedcontainers==2.1.0']

extras_require = \
{'cirq': ['cirq == 0.5.0'],
 'ionq': ['requests >= 2.22.0'],
 'pyquil': ['pyquil==2.16.0'],
 'qiskit': ['qiskit-aer >= 0.5.1']}

setup(name='qcware-quasar',
      version='1.0.0',
      description='Build and execute quantum circuits on various backends.',
      author='Rob Parrish',
      author_email='rob.parrish@qcware.com',
      url='http://github.com/qcware/quasar',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )



import numpy as np
import time
from .options import Options

# ==> VQA Optimizers <== #

class Optimizer(object):

    pass

class BFGSOptimizer(Optimizer):

    @staticmethod
    def default_options():
        
        if hasattr(BFGSOptimizer, '_default_options'): return BFGSOptimizer._default_options.copy()
        opt = Options() 

        opt.add_option(
            key='maxiter',
            value=100,
            allowed_types=[int],
            doc='Quantum simulator or hardware backend')
        opt.add_option(
            key='g_convergence',
            value=1.0E-7,
            allowed_types=[float],
            doc='Maximum gradient element criteria for convergence')
            
        BFGSOptimizer._default_options = opt
        return BFGSOptimizer._default_options.copy()

    def __init__(
        self,
        options,
        ):

        """ BFGSOptimizer initialization - no computational effort performed. 
        """
        
        self.options = options

        
    def from_options(**kwargs):
        """ Return an instance of this class with default options updated from values in kwargs. """
        return BFGSOptimizer(BFGSOptimizer.default_options().set_values(kwargs))

    def optimize(
        self,
        observable,
        guess_parameters,
        print_level=1,
        ):

        if print_level:
            print(' > BFGS Optimizer <\n')

        if print_level:
            print('  %-13s = %d' % ('maxiter', self.options['maxiter']))
            print('  %-13s = %.3E' % ('g_convergence', self.options['g_convergence']))
            print('')

        Z = np.array(guess_parameters)

        import scipy.optimize
        Z, sa_E, opt_data = scipy.optimize.fmin_l_bfgs_b(
            x0=Z,
            func=observable.run_observable,
            fprime=observable.run_observable_gradient,
            iprint=print_level > 0,   
            maxiter=self.options['maxiter'],
            pgtol=self.options['g_convergence'],
            factr=1.0E-6, # Prevent energy difference from being used as stopping criteria
            )
        Z = np.array(Z)

        if print_level:
            print(' > End BFGS Optimizer <\n')

        return Z



def format_statevector(
    statevector,
    real=False,
    formatstr='%+9.6f',
    cutoff=1.0E-6,
    ):

    """ Format a statevector for pretty printing, returning the resulting str.

    Params:
        statevector (np.ndarray of shape (2**N,)) - the statevector to format
        real (bool) - print only the real part of the amplitudes (False) or the
            full complex amplitudes (True - default).
        formatstr (str) - the format str for the real or imaginary part of each
            amplitude.
        cutoff (float) - amplitudes with abs less than cutoff will not be
            printed (set to 0.0 to print all amplitudes).
    Returns:
        (str) - a multiline string with the amplitudes and ket strings
            formatted nicely
    """

    N = (statevector.shape[0]&-statevector.shape[0]).bit_length()-1

    if real:
        formatstr2 = '%s' % (formatstr)
    else:
        formatstr2 = '+(%s%sj)' % (formatstr, formatstr)
    formatstr2 += '|%s>'

    lines = []
    for k, v in enumerate(statevector):
        # Cutoff
        if abs(v) < cutoff: continue
        # Ket str
        start = bin(k)[2:]
        padded = ('0' * (N - len(start))) + start
        # Strval
        if real:
            strval = formatstr2 % (v.real, padded)
        else:
            strval = formatstr2 % (v.real, v.imag, padded)
        lines.append(strval)

    return '\n'.join(lines)
        



import collections

class Option(object):

    """ Class Option represents a key, value Option, with possible restrictions
        on type and value, and with documentation.
    """

    def __init__(
        self,
        key=None,
        value=None,
        required=False,
        allowed_types=None,
        allowed_values=None,
        doc="", 
        ):

        """ Option constructor:

        Params/Members:
            key - the string value of the key of the option.
            value - the value of the option.
            required - True if the user is required to specify the option,
                False if the default value is acceptable. If required is True,
                an error will be raised when get_value is called if value is
                None. 
            allowed_types - array of allowed types of value. If this field is
                not None, a check will be performed on each call to set_value
                and get_value to ensure that value isinstance of at least one
                of the allowed_types.
            allowed_values - list of allowed values of value. If this field is
                not None, a check will be performed on each call to set_value 
                and get_value to ensure the value is one of the allowed_values.
            doc - string message providing helpful documentation about the
                option.
        """
    
        self.key = key
        self.value = value
        self.required = required
        self.allowed_types = allowed_types
        self.allowed_values = allowed_values
        self.doc = doc

    def get_value(self):

        """ Get value for this Option and check validity.

        Returns:
            value if the Option is in a valid state, else raises RuntimeError.
        """
        
        if self.required and self.value is None:
            raise RuntimeError("Option %s is required" % self.key)

        return self.value

    def set_value(self, value):

        """ Set value for this Option and check validity.

        Result:
            value is updated if Option is valid, else raises RuntimeError.
        """

        # Short-circuit if the value is None and the option is not required
        if value is None and not self.required:
            self.value = value
            return

        if self.allowed_types and not any(isinstance(value, x) for x in self.allowed_types):
            raise RuntimeError("Option %s must be one of allowed types: %s" % (self.key, self.allowed_types))
        if self.allowed_values and value not in self.allowed_values:
            raise RuntimeError("Option %s must be one of allowed values: %r" % (self.key, self.allowed_values))

        self.value = value

    def __str__(self):

        """ Return a string containing the full contents and documentation of this Option. """

        s = ''
        s += 'Option:\n'
        s += '  Key: %s\n' % self.key
        s += '  Value: %s\n' % self.value
        s += '  Required: %s\n' % self.required
        s += '  Allowed Types: %s\n' % self.allowed_types
        s += '  Allowed Values: %s\n' % self.allowed_values
        s += '  Doc: %s\n' % self.doc
        s += '\n'
        return s

class Options(object):

    """ Class Options represents a dict of key, value Option objects, including
        restrictions on type, value, etc. Users should interact only with the
        Options class - the Option class is used as internal data storage.

        Generally, codes declaring and using Options objects should first
        define the valid options, rules, and defaults by using the "add_option"
        method. Then, when the user wishes to set or get the values of specific
        options, a copy of the Options object should be provided for the user
        by calling the "copy" method of Options. 

        The underlying Option objects are stored in the options field, which is
        presently a collections.OrderedDict to remember the order of Option
        declaration. This incurs a 2x performance penalty vs. a standard dict
        object, so we may want to optimize the performance later.
    """

    def __init__(
        self,
        options=None,
        ):

        """ Options constructor.

        Params/Members:
            - options - dict of key -> Option
        """

        if options is None: self.options = collections.OrderedDict()
        else: self.options = options

    def add_option(
        self,
        **kwargs
        ):

        """ Declare a new Option with possible default value, type and value
            rules, and documentation.

            Params: See Option constructor for valid kwargs
            Result: Options updated with new Option corresponding to key 
        """

        self.options[kwargs['key']] = Option(
            **kwargs
            )
    
    def get_option(
        self,
        key,
        ):

        """ Get the Option corresponding to key (useful for doc searching an debugging).

        Params:
            - key - string key of Option (raises RuntimeError if not in Options)
        Returns:
            - option - the explicit Option object (most users instead want the
              *value* of this object, which should be accessed through the
              __getitem__ method below).
        """

        if key not in self.options:
            raise ValueError("Key %s is not in Options" % key)
        return self.options[key]

    def __getitem__(
        self,
        key,
        ):

        """ Get the current value of Option corresponding to key, performing validity checks.

        Params:
            - key - string key of Option (raises RuntimeError if not in Options)
        Returns:
            - value - value of Option (raises RuntimeError if type, value or other validity error).
        """
        
        if key not in self.options:
            raise ValueError("Key %s is not in Options" % key)
        return self.options[key].get_value()

    def __setitem__(
        self,
        key,
        value,
        ):

        """ Set the value of Option corresponding to key, performing validity checks.

        Params:
            - key - string key of Option (raises RuntimeError if not in Options)
            - value - value of Option (raises RuntimeError if type, value or other validity error).
        Result:
            - Option value is updated if valid.
        """

        if key not in self.options:
            raise ValueError("Key %s is not in Options" % key)
        return self.options[key].set_value(value)

    def set_values(
        self,
        options,
        ):

        """ Set the values of multiple options. 

        Params:
            - options - dict of key, value pairs to set (calls __setitem__ once
              per key, value pair).
        Results:
            - Option values are updated if valid.
        """
        
        for k, v in options.items():
            self[k] = v
        return self

    def copy(self): 

        """ Return a 1-level shallow copy of this Options object. This makes
            copies of all underlying Option objects so that changes to the new
            Options object will not affect the original Options object.
        """

        options2 = collections.OrderedDict()
        for k, v in self.options.items():
            options2[k] = Option(**v.__dict__)
        return Options(options=options2)

    def __str__(self):

        """ Return the string representations of all Option objects in this Options, in insertion order. """
    
        s = ''.join(str(v) for v in list(self.options.values()))
        return s
        
